from __future__ import annotations

import warnings

from pulser.register.register_layout import RegisterLayout
from qadence2_ir.types import Model

from .device_settings import Fresnel1Settings
from .._base_analog.register import RegisterTransform, RegisterResolver

warnings.filterwarnings("ignore", category=UserWarning)
warnings.formatwarning = lambda msg, *args, **kwargs: f"WARNING: {msg}\n"


def from_model(model: Model) -> RegisterLayout:
    register = RegisterResolver.resolve_from_model(
        model=model,
        device_settings=Fresnel1Settings,
        grid_scale_fn=check_grid_scale,
        grid_type_fn=check_grid_type,
        directives_fn=check_directives,
        register_transform_fn=from_layout,
    )

    return register


def from_layout(register_transform: RegisterTransform) -> RegisterLayout:
    """
    Transforms a data from coordinates into a register with calibrated layout.

    Args:
        register_transform (RegisterTransform): The instance of RegisterTransform containing
            the coordinates and device settings

    Returns:
        A register data with calibrated layout
    """

    layout = register_transform.get_calibrated_layout("TriangularLatticeLayout(61, 5.0µm)")
    traps = layout.get_traps_from_coordinates(*register_transform.coords)
    register = layout.define_register(*traps, qubit_ids=range(len(traps)))

    return register  # type: ignore


def check_grid_scale(model: Model) -> None:
    if Fresnel1Settings.scale_in_range(model.register.grid_scale):
        warnings.warn(
            "Currently, Fresnel-1 uses a fixed grid.",
            SyntaxWarning,
            stacklevel=2,
        )


def check_grid_type(model: Model) -> None:
    if (
        model.register.grid_type
        and model.register.grid_type not in Fresnel1Settings.available_grid_types
    ):
        warnings.warn(
            "Fresnel-1 only supports triangular grids at the moment.",
            SyntaxWarning,
            stacklevel=2,
        )
        model.register.grid_type = Fresnel1Settings.available_grid_types[0]


def check_directives(model: Model) -> None:
    if model.directives.get("enable_digital_analog"):
        warnings.warn(
            "Fresnel-1 uses a fixed grid and does not have digital channels.\n"
            + "Digital operations using atomic distance-based strategies\n"
            + "may not behave as expected.",
            SyntaxWarning,
            stacklevel=2,
        )



from __future__ import annotations

from typing import Union, Callable, Any

import numpy as np
from numpy.typing import ArrayLike
from pulser import AnalogDevice
from pulser.devices import Device
from pulser.register import RegisterLayout
from qadence2_ir.types import Model

from qadence2_platforms.backends._base_analog.device_settings import DeviceSettings
from qadence2_platforms.backends.utils import gridtype_literal

qubits_pos_type = list[tuple[int, int]]
coords_type = Union[ArrayLike, list[ArrayLike], tuple[ArrayLike]]


class RegisterTransform:
    """Transforms register data according to the `grid_type` in the `qadence2_ir.types.Model`"""

    _device: Device
    _grid: gridtype_literal
    _grid_scale: float
    _raw_coords: coords_type
    coords: coords_type

    # below is a multiplying factor that seems to be needed to correct the values coming
    # from the `qadence2-expressions`'s `qubit_positions`. It should be used on the
    # coordinate transformations
    scale_factor: int = 5

    def __init__(
        self,
        device_settings: DeviceSettings,
        grid_transform: gridtype_literal | None,
        grid_scale: float = 1.0,
        coords: qubits_pos_type | None = None,
        num_qubits: int | None = None,
    ):
        """
        Args:
            grid_transform (Literal["linear", "triangular", "square"], None): literal str to choose
                which grid transform to use. Accepted values are "linear", "triangular"
                or "square". If None is provided, it will default to "triangular"
            grid_scale (float): scale of the grid. Default is `1.0`
            coords (list[tuple[int, int]]): list of coordinates as qubit positions in an int
                grid, e.g. `[(0, 0), (1, 0), (0, 1)]`. Default is `None`
            num_qubits (int | None): number of qubits as integer. Default is `None`
            device_settings (DeviceSettings | None): Device settings object
        """

        self._grid = grid_transform if grid_transform is not None else "triangular"
        self._grid_scale = grid_scale

        if coords:
            self._raw_coords = coords

        elif num_qubits:
            self._raw_coords = self._fill_coords(num_qubits)

        else:
            raise ValueError("must provide coords or num_qubits.")

        try:
            self.coords = getattr(self, f"{self._grid}_coords")()
        except AttributeError:
            self.invalid_grid_value()
        else:
            self._device = device_settings.device

    @property
    def raw_coords(self) -> coords_type:
        return self._raw_coords

    @property
    def device(self) -> Device:
        return self._device

    @classmethod
    def _fill_coords(cls, num_qubits: int) -> coords_type:
        shift = num_qubits // 2
        return [(p - shift, 0) for p in range(num_qubits)]

    @classmethod
    def invalid_grid_value(cls) -> None:
        """Fallback function for invalid `grid_transform` value."""

        raise ValueError("grid_transform should be 'linear', 'triangular', or 'square'.")

    def linear_coords(self) -> np.ndarray:
        """
        Transforms coordinates into linear coordinates.

        Returns:
            np.ndarray of transformed coordinates.
        """

        raise NotImplementedError()

    def triangular_coords(self) -> np.ndarray:
        """
        Transforms coordinates into triangular coordinates.

        Returns:
            np.ndarray of transformed coordinates
        """

        # triangular transformation matrix
        transform = np.array([[1.0, 0.0], [0.5, 0.8660254037844386]])
        return np.array(self._raw_coords) * self._grid_scale * self.scale_factor @ transform

    def square_coords(self) -> np.ndarray:
        """
        Transforms coordinates into square coordinates.

        Returns:
            np.ndarray of transformed coordinates
        """

        # for now, no transformation needed since the coords are list of tuple of ints
        return np.array(self._raw_coords) * self._grid_scale * self.scale_factor

    @classmethod
    def get_calibrated_layout(cls, layout_name: str) -> RegisterLayout:
        """
        Gets the calibrated layout for the given `layout_name` according to `AnalogDevice`
        specifications.

        Args:
            layout_name: the name of the layout.

        Returns:
            The `RegisterLayout` object for the given `layout_name`.
        """

        return AnalogDevice.calibrated_register_layouts.get(layout_name, None)


class RegisterResolver:
    """
    An object to hold common functionalities for registers and resolving their
    layouts, transformations and checks.
    """

    @classmethod
    def resolve_from_model(
        cls,
        model: Model,
        device_settings: DeviceSettings,
        grid_scale_fn: Callable[[Model], Any],
        grid_type_fn: Callable[[Model], Any],
        directives_fn: Callable[[Model], Any],
        register_transform_fn: Callable[[RegisterTransform], RegisterLayout],
    ) -> RegisterLayout:
        """
        Resolves the model's register data into actual platform appropriate and
        validated data through platform's own functions. It is a generic and
        modular helper function to be called by `register.from_model` function.

        Args:
            model (Model): The model to use and resolve its data
            device_settings (DeviceSettings): The device settings instance
            grid_scale_fn (Callable): Function used to check the grid scale
                against the device's
            grid_type_fn (Callable): Function used to check the grid type
                against the device's
            directives_fn (Callable): Function used to check the directives
                against the device's
            register_transform_fn (Callable): Function used to transform
                coordinates into register on appropriate layout

        Returns:
            A register data with appropriate layout
        """

        model_register = model.register

        if model_register.num_qubits < 1:
            raise ValueError("No qubit available in the register.")

        grid_scale_fn(model)
        grid_type_fn(model)
        directives_fn(model)

        register_transform = RegisterTransform(
            grid_transform=model_register.grid_type,
            grid_scale=model_register.grid_scale,
            coords=model_register.qubit_positions,
            num_qubits=model_register.num_qubits,
            device_settings=device_settings,
        )

        register = register_transform_fn(register_transform)

        return register



from __future__ import annotations

from logging import getLogger
from typing import Any, Counter, Iterable, Literal, cast

import pyqtorch as pyq
import torch
from pyqtorch.utils import DiffMode
from torch.nn import ParameterDict

from qadence2_platforms.abstracts import (
    AbstractInterface,
    RunEnum,
)
from qadence2_platforms.backends.utils import InputType

from .embedding import Embedding
from .functions import parse_native_observables
from .register import RegisterInterface

logger = getLogger(__name__)


class Interface(
    AbstractInterface[
        torch.Tensor,
        pyq.QuantumCircuit,
        torch.Tensor,
        torch.Tensor,
        list[Counter],
        torch.Tensor,
    ],
):
    """A class holding register, embedding, circuit, native backends and optional observable."""

    def __init__(
        self,
        register: RegisterInterface,
        embedding: Embedding,
        circuit: pyq.QuantumCircuit,
        vparams: dict[str, torch.Tensor] = None,
        observable: list[InputType] | InputType | None = None,
    ) -> None:
        super().__init__()
        self.register = register
        self.init_state: torch.Tensor = (
            circuit.state_from_bitstring(register.init_state)
            if register.init_state is not None
            else circuit.init_state()
        ).to(dtype=torch.complex128)
        self.embedding = embedding
        self.circuit = circuit
        self.observable = observable
        self.vparams = ParameterDict(vparams)
        self._dtype = torch.float64

    @property
    def info(self) -> dict[str, Any]:
        return {"num_qubits": self.register.n_qubits}

    @property
    def sequence(self) -> pyq.QuantumCircuit:
        return self.circuit

    def add_noise(self, model: Literal["SPAM"]) -> None:
        pass

    def set_parameters(self, params: dict[str, float]) -> None:
        pass

    def parameters(self) -> Iterable[Any]:
        return self.vparams.values()  # type: ignore [no-any-return]

    def draw(self, values: dict[str, Any]) -> None:
        raise NotImplementedError("PyQTorch currently does not support drawing the circuit")

    def _run(
        self,
        run_type: RunEnum,
        values: dict[str, torch.Tensor] | None = None,
        state: torch.Tensor | None = None,
        shots: int | None = None,
        observable: list[InputType] | InputType | None = None,
        diff_mode: DiffMode = None,
        **_: Any,
    ) -> Any:
        """
        Method to execute run type-specific option.

        Option can be `run`, `sample`
        or `expectation`. Each option can have different arguments, such as:
        `sample` uses `shots`, while `expectation` uses `observable`.

        It should not be called directly. Use it on `run`, `sample` or `expectation`
        methods.

        :param run_type: str option as `run`, `sample` or `expectation`
        :param values: dictionary of user-input parameters
        :param callback: callback function to be used internally, if applicable
        :param state: a tensor containing the desired state to perform the execution from
        :param shots: number of shots, if applicable (`sample` only)
        :param observable: a list of observables, if applicable (`expectation` only)
        :return: a tensor or list of values (`sample` only) of the calculated state
        """

        def set_dtype(data: dict[str, torch.Tensor] | None) -> dict[str, torch.Tensor] | None:
            if data is None:
                return None
            return {k: v.to(dtype=self._dtype) for k, v in data.items()}

        inputs: dict[str, torch.Tensor] = set_dtype(values) or dict()
        state = state.to(dtype=torch.complex128) if state is not None else self.init_state

        match run_type:
            case RunEnum.RUN:
                return pyq.run(
                    circuit=self.circuit,
                    state=state,
                    values=inputs,
                    embedding=self.embedding,
                )
            case RunEnum.SAMPLE:
                return pyq.sample(
                    circuit=self.circuit,
                    state=state,
                    values=inputs,
                    n_shots=shots,
                    embedding=self.embedding,
                )
            case RunEnum.EXPECTATION:
                if observable is not None or self.observable is not None:
                    return pyq.expectation(
                        circuit=self.circuit,
                        state=state,
                        values={**self.vparams, **inputs},
                        observable=parse_native_observables(observable or self.observable),  # type: ignore [arg-type]
                        embedding=self.embedding,
                        diff_mode=diff_mode,
                    )
                raise ValueError("Observable must not be None for expectation run.")
            case _:
                raise NotImplementedError(f"Run type '{run_type}' not implemented.")

    def run(
        self,
        values: dict[str, torch.Tensor] | None = None,
        state: torch.Tensor | None = None,
        **kwargs: Any,
    ) -> torch.Tensor:
        return self._run(RunEnum.RUN, values=values, state=state, **kwargs)

    def sample(
        self,
        values: dict[str, torch.Tensor] | None = None,
        shots: int | None = None,
        state: torch.Tensor | None = None,
        **kwargs: Any,
    ) -> list[Counter]:
        return cast(
            list,
            self._run(
                RunEnum.SAMPLE,
                values=values,
                shots=shots,
                state=state,
                **kwargs,
            ),
        )

    def expectation(
        self,
        values: dict[str, torch.Tensor] | None = None,
        observable: list[InputType] | InputType | None = None,
        state: torch.Tensor | None = None,
        diff_mode: DiffMode = DiffMode.AD,
        **kwargs: Any,
    ) -> torch.Tensor:
        return self._run(
            RunEnum.EXPECTATION,
            values=values,
            state=state,
            observable=observable,
            diff_mode=diff_mode,
            **kwargs,
        )

    def __call__(self, *args: Any, **kwargs: Any) -> torch.Tensor:
        return self.run(*args, **kwargs)



from __future__ import annotations

from dataclasses import dataclass
from logging import getLogger

import pyqtorch as pyq
import torch
from pyqtorch.quantum_operation import QuantumOperation
from qadence2_ir.types import Alloc, Load, Model, QuInstruct

from qadence2_platforms.backends.pyqtorch.embedding import Embedding
from qadence2_platforms.backends.pyqtorch.interface import Interface
from qadence2_platforms.backends.pyqtorch.register import RegisterInterface

logger = getLogger(__name__)


class Compiler:
    """
    A dataclass to compile an IR model data into PyQTorch objects (to be run in a
    PyQTorch-based backend.
    """

    instruction_mapping = {
        "not": pyq.CNOT,
        "add": pyq.Add,
        "mul": pyq.Scale,
        "noncommute": pyq.Sequence,
    }

    @classmethod
    def _get_target(cls, target: tuple[int, ...] | tuple, num_qubits: int) -> tuple[int, ...]:
        return tuple(range(num_qubits)) if len(target) == 0 else target

    def compile(
        self,
        model: Model,
    ) -> pyq.QuantumCircuit:
        """
        Compiling IR model data to PyQTorch object function. It transforms model
        `QuInstruct`s into PyQTorch operators, resolving the SSA-form arguments
        into concrete values or valid PyQTorch parameters.

        Args:
            model (Model): IR model to compile

        Returns:
            A PyQTorch quantum circuit object with the model `QuInstruct`s compiled into
            PyQTorch operators
        """

        pyq_operations = []

        for instr in model.instructions:

            if isinstance(instr, QuInstruct):
                native_op: QuantumOperation = getattr(
                    pyq, instr.name.upper(), self.instruction_mapping.get(instr.name)
                )
                control = instr.support.control
                target = self._get_target(instr.support.target, model.register.num_qubits)
                native_support = (*control, *target)

                if len(instr.args) > 0:
                    assert len(instr.args) == 1, "More than one arg not supported"
                    (maybe_load,) = instr.args
                    assert isinstance(maybe_load, Load), "only support load"
                    pyq_operations.append(
                        native_op(native_support, maybe_load.variable).to(dtype=torch.complex128)
                    )

                else:
                    pyq_operations.append(native_op(*native_support).to(dtype=torch.complex128))

        return pyq.QuantumCircuit(model.register.num_qubits, pyq_operations).to(
            dtype=torch.complex128
        )


def get_trainable_params(inputs: dict[str, Alloc]) -> dict[str, torch.Tensor]:
    return {
        param: torch.rand(value.size, requires_grad=True)
        for param, value in inputs.items()
        if value.is_trainable
    }


def compile_to_backend(model: Model) -> Interface:
    """
    Compiles the model data (IR information from expressions) into PyQTorch-compatible data and
    defines an Interface instance to be available to the user to invoke useful methods, such as
    `run`, `sample`, `expectation`, `set_parameters`.

    Args:
        model (Model): the IR model data to be compiled to PyQTorch-based backend

    Returns:
        The `Interface` instance based on PyQTorch backend
    """

    register_interface = RegisterInterface(
        model.register.num_qubits, model.register.options.get("init_state")
    )
    embedding = Embedding(model)
    native_circ = Compiler().compile(model)
    vparams = get_trainable_params(model.inputs)
    return Interface(register_interface, embedding, native_circ, vparams=vparams)



from __future__ import annotations

from functools import reduce
from typing import Any

import numpy as np
from pulser.parametrized.variable import VariableItem
from pulser.sequence.sequence import Sequence
from qadence2_ir.types import Alloc, Assign, Call, Load, QuInstruct


class NamedPulse:
    def __init__(self, name: str, *args: Any) -> None:
        self.name = name
        self.args = args


def from_instructions(
    sequence: Sequence,
    inputs: dict[str, Alloc],
    instructions: list[Assign | QuInstruct],
    allow_time_dependent: bool = False,
) -> list[NamedPulse]:
    variables = dict()
    temp_vars: dict[str, Any] = dict()

    for var in inputs:
        # inputs[var].size holds the points to interpolate time-dependent functions
        if inputs[var].size > 1:
            if allow_time_dependent:
                variables[var] = sequence.declare_variable(var, size=inputs[var].size)
            else:
                raise TypeError("This platform cannot handle time modulated variables.")
        else:
            variables[var] = sequence.declare_variable(var)

    pulses = []
    for instruction in instructions:
        if isinstance(instruction, Assign):
            assign(instruction, temp_vars, variables)

        if isinstance(instruction, QuInstruct):
            args = (
                (
                    (temp_vars.get(arg.variable) or variables[arg.variable])
                    if isinstance(arg, Load)
                    else arg
                )
                for arg in instruction.args
            )
            pulses.append(NamedPulse(instruction.name, *args))

    return pulses


def assign(
    instruction: Assign, temp_vars: dict[str, Any], variables: dict[str, VariableItem]
) -> None:
    var_name = instruction.variable

    if var_name in temp_vars or var_name in variables:
        return

    fn = instruction.value
    if isinstance(fn, Call):
        args = (
            (
                (temp_vars.get(arg.variable) or variables.get(arg.variable))
                if isinstance(arg, Load)
                else arg
            )
            for arg in fn.args
        )
        temp_vars[var_name] = _compute(fn.identifier, *args)


def _compute(fn: str, *args: Any) -> Any:
    match fn:
        case "add":
            return reduce(lambda a, b: a + b, args)
        case "mul":
            return reduce(lambda a, b: a * b, args)
        case "pow":
            return reduce(lambda a, b: a**b, args)
        case _:
            return getattr(np, fn)(*args)



from __future__ import annotations

from typing import Any, Literal, Protocol, runtime_checkable


gridtype_literal = Literal["linear", "triangular", "square"]


@runtime_checkable
class Support(Protocol):
    def target_all(self) -> Any: ...

    @property
    def target(self) -> list[int]: ...

    @property
    def control(self) -> list[int]: ...

    @property
    def subspace(self) -> set[int]: ...

    @property
    def max_index(self) -> int: ...


@runtime_checkable
class InputType(Protocol):
    @property
    def head(self) -> InputType | str: ...

    @property
    def args(self) -> InputType | list[InputType | str | Support]: ...

    @property
    def value(self) -> Any: ...

    @property
    def is_symbol(self) -> bool: ...

    @property
    def is_quantum_operator(self) -> bool: ...

    @property
    def is_addition(self) -> bool: ...

    @property
    def is_multiplication(self) -> bool: ...

    @property
    def is_kronecker_product(self) -> bool: ...

    @property
    def is_power(self) -> bool: ...

    @property
    def subspace(self) -> Support | None: ...

    def add(self, *args: InputType) -> InputType: ...

    def mul(self, *args: InputType) -> InputType: ...

    def kron(self, *args: InputType) -> InputType: ...

    def pow(self, *args: InputType) -> InputType: ...

    def __getitem__(self, item: slice | int) -> Any: ...



from __future__ import annotations

from typing import Any

import numpy as np
from pulser import Pulse
from pulser.parametrized.variable import VariableItem, Variable
from pulser.sequence import Sequence
from pulser.waveforms import ConstantWaveform, BlackmanWaveform, RampWaveform, CompositeWaveform

from qadence2_platforms.backends._base_analog.functions import (
    Direction,
    Duration,
    base_parse_native_observables,
    BaseQuTiPObservablesParser,
    DEFAULT_AMPLITUDE,
    DEFAULT_DETUNING,
)

# TODO: re-introduce `Support` to account for "local" and "global" on the `channel` arg

# pulse function mapping.
# keys are the `qadence2_ir.Model` standard, values are the function names below
PULSE_FN_MAP = {
    "rx": "rx",
    "ry": "ry",
    "not": "x",
    "h": "h",
}

# reference to function to parse native observables on `Interface`'s `expectation` method
parse_native_observables = base_parse_native_observables
# qutip observable parser
QuTiPObservablesParser = BaseQuTiPObservablesParser


def dyn_pulse(
    sequence: Sequence,
    duration: VariableItem | float,
    amplitude: VariableItem | float,
    detuning: VariableItem | float,
    phase: VariableItem | float,
    **_: Any,
) -> None:
    """
    Dynamic pulse to simulate a specific time-dependent hamiltonian for neutral-atom devices.

    Args:
        sequence: a `pulser.sequence.Sequence` instance
        duration: duration of the pulse in dimensionless units
        amplitude: amplitude of the pulse in dimensionless units
        detuning: detuning of the pulse in dimensionless units
        phase: phase in radians
    """
    max_amp = sequence.device.channels["rydberg_global"].max_amp or DEFAULT_AMPLITUDE
    max_abs_detuning = (
        sequence.device.channels["rydberg_global"].max_abs_detuning or DEFAULT_DETUNING
    )

    # FIXME: Centralize unit converions
    duration *= 1000 * 2 * np.pi / max_amp  # type: ignore
    amplitude *= max_amp  # type: ignore
    detuning *= max_abs_detuning  # type: ignore

    new_amplitude = ConstantWaveform(duration, amplitude)
    new_detuning = ConstantWaveform(duration, detuning)

    p = Pulse(new_amplitude, new_detuning, phase)
    sequence.add(p, channel="global")


def piecewise_pulse(
    sequence: Sequence,
    duration: Variable | VariableItem,
    amplitude: Variable,
    detuning: Variable,
    phase: VariableItem | float,
    **_: Any,
) -> None:
    """
    Dynamic pulse to simulate a specific piecewise time-dependent hamiltonian for neutral-atom devices.

    Args:
        sequence: a `pulser.sequence.Sequence` instance
        duration: duration of the pulse in dimensionless units
        amplitude: amplitude of the pulse in dimensionless units
        detuning: detuning of the pulse in dimensionless units
        phase: phase in radians
    """
    max_amp = sequence.device.channels["rydberg_global"].max_amp or DEFAULT_AMPLITUDE
    max_abs_detuning = (
        sequence.device.channels["rydberg_global"].max_abs_detuning or DEFAULT_DETUNING
    )

    dur_factor = 1000 * 2 * np.pi / max_amp  # type: ignore
    amp_factor = max_amp  # type: ignore
    det_factor = max_abs_detuning  # type: ignore

    # Needed so a size = 1 variable is made iterable
    duration = [duration] if isinstance(duration, VariableItem) else duration

    for i, dur in enumerate(duration):
        amp_wf = RampWaveform(
            dur * dur_factor, amplitude[i] * amp_factor, amplitude[i + 1] * amp_factor
        )
        det_wf = RampWaveform(
            dur * dur_factor, detuning[i] * det_factor, detuning[i + 1] * det_factor
        )
        sequence.add(Pulse(amp_wf, det_wf, phase), "global")


def rx(
    sequence: Sequence,
    angle: VariableItem | float,
    **_: Any,
) -> None:
    rotation(sequence, angle, Direction.X)


def ry(
    sequence: Sequence,
    angle: VariableItem | float,
    **_: Any,
) -> None:
    rotation(sequence, angle, Direction.Y)


def x(sequence: Sequence, **_: Any) -> None:
    rotation(sequence, angle=np.pi, direction=Direction.X)


def h(
    sequence: Sequence,
    duration: VariableItem | float = np.pi,
    support_list: str = "global",
    **_: Any,
) -> None:
    amplitude = sequence.device.channels["rydberg_global"].max_amp or DEFAULT_AMPLITUDE
    duration *= 1000 * 2 * np.pi / amplitude

    pi2_wf = BlackmanWaveform(1000, np.pi / 2)
    sequence.add(
        Pulse.ConstantDetuning(pi2_wf, 0, np.pi / 2, post_phase_shift=np.pi), channel="global"
    )


def rotation(
    sequence: Sequence,
    angle: VariableItem | float,
    direction: Direction | float,
    **_: Any,
) -> None:

    amplitude = sequence.device.channels["rydberg_global"].max_amp or DEFAULT_AMPLITUDE
    duration = 1000 * angle / amplitude
    detuning = 0

    phase: Any
    match direction:
        case Direction.X:
            phase = 0
        case Direction.Y:
            phase = np.pi / 2
        case _:
            phase = direction

    sequence.add(Pulse.ConstantPulse(duration, amplitude, detuning, phase), "global")


def dyn_wait(
    sequence: Sequence,
    duration: VariableItem | float,
    *args: Any,
    **kwargs: Any,
) -> None:
    max_amp = sequence.device.channels["rydberg_global"].max_amp or DEFAULT_AMPLITUDE
    duration *= 1000 * 2 * np.pi / max_amp  # type: ignore
    sequence.delay(int(duration), "global")  # type: ignore


def apply_local_shifts(sequence: Sequence, **_: Any) -> None:
    max_abs_detuning = (
        sequence.device.channels["rydberg_global"].max_abs_detuning or DEFAULT_DETUNING
    )
    time_scale = 1000 * 2 * np.pi / max_abs_detuning
    local_pulse_core(sequence, duration=1.0, time_scale=time_scale, detuning=1.0, concurrent=False)


def local_pulse(
    sequence: Sequence,
    duration: VariableItem | float | Duration,
    detuning: VariableItem | float,
    concurrent: bool = False,
    **_: Any,
) -> None:
    max_amp = sequence.device.channels["rydberg_global"].max_amp or DEFAULT_AMPLITUDE
    time_scale = 1000 * 2 * np.pi / max_amp
    local_pulse_core(sequence, duration, time_scale, detuning, concurrent)


def local_pulse_core(
    sequence: Sequence,
    duration: VariableItem | float | Duration,
    time_scale: float,
    detuning: VariableItem | float,
    concurrent: bool = False,
    **kwargs: Any,
) -> None:
    max_abs_detuning = (
        sequence.device.channels["rydberg_global"].max_abs_detuning or DEFAULT_DETUNING
    )

    if duration == Duration.FILL:
        if not concurrent:
            raise SyntaxError("The option `fill` can only be used on the `concurrent` mode")

        duration = sequence.get_duration("global") - sequence.get_duration("dmm_0")

    else:
        duration *= time_scale  # type: ignore

    detuning *= -1 * max_abs_detuning  # type: ignore

    sequence.add_dmm_detuning(
        ConstantWaveform(duration, detuning),
        "dmm_0",
        "no-delay" if concurrent else "min-delay",
    )



from __future__ import annotations

from typing import Callable, Optional

import numpy as np
from pulser.register.register_layout import RegisterLayout
from pulser.sequence.sequence import Sequence
from qadence2_ir.types import Model

from qadence2_platforms.backends._base_analog.sequence import from_instructions

from . import functions as add_pulse
from .device_settings import Fresnel1Settings
from .functions import PULSE_FN_MAP


# Fresnel-1 device (AnalogDevice with Fresnel-1 specs)
Fresnel1 = Fresnel1Settings.device


def from_model(model: Model, register: RegisterLayout) -> Sequence:
    seq = Sequence(register, Fresnel1)  # type: ignore
    seq.declare_channel("global", "rydberg_global")

    if model.directives.get("local_targets"):
        if model.directives.get("local_shifts"):
            targets = zip(
                model.directives["local_targets"],
                model.directives["local_shifts"],
            )
            detuning_map = register.define_detuning_map(
                {elem[0]: elem[1] / (2 * np.pi) for elem in targets}
            )
        else:
            targets = zip(
                model.directives["local_targets"],
                [1] * len(model.directives["local_targets"]),
            )
            detuning_map = register.define_detuning_map(dict(targets))

        seq.config_detuning_map(detuning_map, "dmm_0")

    pulses = from_instructions(seq, model.inputs, model.instructions)

    for pulse in pulses:
        fn: Optional[Callable] = getattr(
            add_pulse, PULSE_FN_MAP.get(pulse.name) or pulse.name, None
        )
        if fn is not None:
            fn(seq, *pulse.args)
        else:
            raise ValueError(f"current backend does not have pulse '{pulse.name}' implemented.")

    return seq



from __future__ import annotations

from abc import ABC

from pulser.devices import Device


class DeviceSettings(ABC):
    """
    Device settings to ease the building of sequence, register and interface logic

    name (str): the name of the device, ex: `Analog Device`
    name_short (str): a shorter name version, ex: `analog`
    device (pulser.devices.Device): the pulser device
    grid_scale_range (tuple[float, float]): a tuple of min and max values for the
        grid scale, ex: `(1.0, 1.0)` (should not scale), `(1.0, 10.0)` (up to
        10.0, included)
    available_grid_types (tuple[str]): a tuple of all the possible grid types
        for the device, ex: `("triangular",)`, `("linear", "square")`
    available_directives (tuple[str], optional): a tuple of available directives,
        ex: `("enable_digital_analog")`
    """

    _name: str
    _name_short: str
    _device: Device
    _grid_scale_range: tuple[float, float]
    _available_grid_types: tuple[str, ...]
    _available_directives: tuple[str, ...] | tuple | None

    @property
    def name(self) -> str:
        return self._name

    @property
    def name_short(self) -> str:
        return self._name_short

    @property
    def device(self) -> Device:
        return self._device

    @property
    def grid_scale_range(self) -> tuple[float, float]:
        return self._grid_scale_range

    @property
    def available_grid_types(self) -> tuple[str, ...]:
        return self._available_grid_types

    @property
    def available_directives(self) -> tuple[str, ...] | tuple | None:
        return self._available_directives

    def scale_in_range(self, grid_scale: float) -> bool:
        """
        Check whether the grid scale is within device's range.

        Args:
            grid_scale (float): the grid scale. For a normal grid, the scale is 1.0

        Returns:
            A boolean to whether the grid scale is within device's range.
        """

        return self._grid_scale_range[0] <= grid_scale <= self._grid_scale_range[1]



from __future__ import annotations

import warnings

from pulser.register import RegisterLayout, Register
from qadence2_ir.types import Model

from qadence2_platforms.backends._base_analog.register import RegisterTransform, RegisterResolver
from qadence2_platforms.backends.analog.device_settings import AnalogSettings

warnings.filterwarnings("ignore", category=UserWarning)
warnings.formatwarning = lambda msg, *args, **kwargs: f"WARNING: {msg}\n"


def from_model(model: Model) -> RegisterLayout:
    """
    Gets information from IR model data to generate a register/register layout.

    Args:
        model (Model): IR model data

    Returns:
        `RegisterLayout` data
    """

    register = RegisterResolver.resolve_from_model(
        model=model,
        device_settings=AnalogSettings,
        grid_scale_fn=check_grid_scale,
        grid_type_fn=check_grid_type,
        directives_fn=check_directives,
        register_transform_fn=from_coords,
    )

    return register


def from_coords(register_transform: RegisterTransform) -> RegisterLayout:
    """
    Function to transform the coordinates into appropriate register layout.

    Args:
        register_transform (RegisterTransform): Register transform instance containing
            the coordinates and register-related data

    Returns:
        `RegisterLayout` data
    """
    register = Register.from_coordinates(register_transform.coords)
    return register  # type: ignore


def check_grid_scale(model: Model) -> None:
    pass


def check_grid_type(model: Model) -> None:
    pass


def check_directives(model: Model) -> None:
    pass



from __future__ import annotations

from dataclasses import replace
import warnings

import numpy as np
from pulser.channels import DMM
from pulser.devices import AnalogDevice

from qadence2_platforms.backends._base_analog.device_settings import DeviceSettings

warnings.filterwarnings("ignore", category=UserWarning)
warnings.formatwarning = lambda msg, *args, **kwargs: f"WARNING: {msg}\n"


class Fresnel1DeviceSettings(DeviceSettings):
    """Defines Fresnel-1 specific settings for checks."""

    def __init__(self) -> None:
        self._name = "Fresnel-1"
        self._name_short = "fresnel1"
        self._device = replace(
            AnalogDevice.to_virtual(),
            dmm_objects=(
                DMM(
                    # from Pulser tutorials/dmm.html#DMM-Channel-and-Device
                    clock_period=4,
                    min_duration=16,
                    max_duration=2**26,
                    mod_bandwidth=8,
                    bottom_detuning=-2 * np.pi * 20,  # detuning between 0 and -20 MHz
                    total_bottom_detuning=-2 * np.pi * 2000,  # total detuning
                ),
            ),
        )
        self._grid_scale_range = (1.0, 1.0)
        self._available_grid_types = ("triangular",)
        # TODO: check which directives should or must be present
        self._available_directives = ()


# Define the Fresnel-1 settings, including the device object and
# its limitations or parameters range.
Fresnel1Settings = Fresnel1DeviceSettings()



from __future__ import annotations

from pulser.devices import AnalogDevice

from qadence2_platforms.backends._base_analog.device_settings import DeviceSettings


class AnalogDeviceSettings(DeviceSettings):
    """Defines AnalogDevice specific settings for checks"""

    def __init__(self) -> None:
        self._name = "AnalogDevice"
        self._short_name = "analog"
        self._device = AnalogDevice
        self._grid_scale_range = (1.0, 100.0)
        # TODO: validate whether the "square" grid type is working properly
        self._available_grid_types = ("triangular", "square")
        self._available_directives = ("local_shifts", "local_targets", "enable_digital_analog")


# Define the AnalogDevice settings, including device object and
# its limitations or parameters range.
AnalogSettings = AnalogDeviceSettings()



from __future__ import annotations

from enum import Enum, auto
from functools import reduce
from typing import Iterable, cast

import numpy as np
import qutip

from qadence2_platforms.backends.utils import InputType, Support

DEFAULT_AMPLITUDE = 4 * np.pi
DEFAULT_DETUNING = 10 * np.pi

# TODO: re-introduce `Support` to account for "local" and "global" on the `channel` arg


class Direction(Enum):
    X = auto()
    Y = auto()


class Duration(Enum):
    FILL = auto()


###############
# OBSERVABLES #
###############


def base_parse_native_observables(
    num_qubits: int, observable: list[InputType] | InputType
) -> list[qutip.Qobj]:
    """
    Function to be called by `Interface`'s `expectation` method on Pulser-based backends
    using QuTiP emulator.

    Args:
        num_qubits (int): number of qubits
        observable (list[InputType] | InputType): the input expression. Any
            qadence2-expressions expression compatible object, with the same
            methods, or a list of it

    Returns:
        A QuTiP object with the Hilbert space compatible with `num_qubits`
    """
    return BaseQuTiPObservablesParser.build(num_qubits, observable)


class BaseQuTiPObservablesParser:
    """
    Convert InputType object to Qutip native quantum objects for simulation on QuTiP.

    It is intended to be used on the expectation method of Pulser-based interface classes.
    InputType can be qadence2-expressions expression or any other module with the same
    methods.
    """

    operators_mapping = {
        "I": qutip.qeye(2),
        "Z": qutip.sigmaz(),
    }

    @classmethod
    def _compl_tensor_op(cls, num_qubits: int, expr: InputType) -> qutip.Qobj:
        """
        Use it for pure kron operations or for single input operator that needs to match
        a bigger Hilbert space, e.g. `expr = Z(0)` but the number of qubits is 3.

        Args:
            num_qubits (int): the number of qubits to create the qutip object to
            expr (InputType): the input expression. Any qadence2-expressions expression
                compatible object, with the same methods

        Returns:
            A QuTiP object with the Hilbert space compatible with `num_qubits`
        """

        op: qutip.Qobj
        arg: InputType

        if expr.subspace:
            native_ops: list[qutip.Qobj] = []
            support_set: set = expr.subspace.subspace

            for k in range(num_qubits):
                if k in support_set:
                    sub_num_qubits: int = cast(Support, expr.args[1]).max_index
                    arg = cast(InputType, expr.args[0])
                    op = cls._get_op(sub_num_qubits, arg)

                else:
                    op = cls.operators_mapping["I"]

                native_ops.append(op)

            return qutip.tensor(*native_ops)

        arg = cast(InputType, expr.args[0])
        op = cls._get_op(num_qubits, arg)
        return qutip.tensor(*([op] * num_qubits))

    @classmethod
    def _arith_tensor_op(cls, num_qubits: int, expr: InputType) -> qutip.Qobj:
        """
        Use it for the arithmetic operations addition and multiplication that need to
        have an extended Hilbert space compatible with `num_qubits`.

        Args:
            num_qubits (int): the number of qubits to create the qutip object to
            expr (InputType): the input expression. Any qadence2-expressions expression
                compatible object, with the same methods

        Returns:
            A QuTiP object with the Hilbert space compatible with `num_qubits`
        """

        subspace: set = cast(Support, expr.subspace).subspace
        super_space: set = set(range(num_qubits))

        if super_space.issuperset(subspace):
            native_ops: list[qutip.Qobj] = []
            arg_subspace: set = cast(Support, expr.args[1]).subspace

            for k in range(num_qubits):
                if k in arg_subspace:
                    sub_num_qubits: int = cast(int, expr.args[1])
                    arg: InputType = cast(InputType, expr.args[0])
                    op: qutip.Qobj = cls._get_op(sub_num_qubits, arg)
                    native_ops.append(op)
                else:
                    native_ops.append(cls.operators_mapping["I"])

            return qutip.tensor(*native_ops)

        raise ValueError(
            f"subspace of the object ({subspace}) is bigger than the "
            f"sequence space ({super_space})"
        )

    @classmethod
    def _get_op(cls, num_qubits: int, op: InputType) -> qutip.Qobj | None:
        """
        Convert an expression into a native QuTiP object. A simple symbol,
        a quantum operator, and an operation (addition, multiplication or
        kron tensor) are valid objects.

        Args:
            num_qubits (int): the number of qubits to create the qutip object to
            op (InputType): the input expression. Any qadence2-expressions expression
                compatible object, with the same methods

        Returns:
            A QuTiP object with the Hilbert space compatible with `num_qubits`
        """

        if op.is_symbol is True:
            symbol: str = cast(str, op.args[0])
            return cls.operators_mapping[symbol]

        op_arg: qutip.Qobj

        if op.is_quantum_operator is True:
            sub_num_qubits: int = cast(Support, op.args[1]).max_index + 1

            if sub_num_qubits < num_qubits:
                op_arg = cls._compl_tensor_op(num_qubits, op)

            else:
                arg: InputType = cast(InputType, op.args[0])
                op_arg = cls._get_op(num_qubits, arg)

            return op_arg

        ops: list[qutip.Qobj] = []
        args: Iterable = cast(Iterable, op.args)

        if op.is_addition is True:

            for arg in args:
                ops.append(cls._arith_tensor_op(num_qubits, arg))

            return reduce(lambda a, b: a + b, ops)

        if op.is_multiplication is True:

            for arg in args:
                ops.append(cls._arith_tensor_op(num_qubits, arg))

            return reduce(lambda a, b: a * b, ops)

        if op.is_kronecker_product is True:
            return cls._compl_tensor_op(num_qubits, op)

        raise NotImplementedError(
            f"could not retrieve the expression {op} ({type(op)}) from the observables"
        )

    @classmethod
    def _iterate_over_obs(cls, n_qubits: int, op: Iterable | InputType) -> list[qutip.Qobj]:
        if isinstance(op, Iterable):
            return [cls._get_op(n_qubits, arg) for arg in op]

        args: Iterable = cast(Iterable, op.args)
        return [cls._get_op(n_qubits, cast(InputType, arg)) for arg in args]

    @classmethod
    def build(cls, num_qubits: int, observables: list[InputType] | InputType) -> list[qutip.Qobj]:
        """
        Parses an input expression or list of expressions into a native QuTiP object.

        Args:
            num_qubits (int): the number of qubits to create the qutip object to
            observables (list[InputType], InputType): the input expression. Any
                qadence2-expressions expression compatible object, with the same
                methods

        Returns:
            A QuTiP object with the Hilbert space compatible with `num_qubits`
        """
        if not isinstance(observables, list):
            return [cls._get_op(num_qubits, observables)]
        return cls._iterate_over_obs(num_qubits, observables)



from __future__ import annotations

from typing import Iterable, cast

from pyqtorch.primitives import Primitive
from torch.nn import Module
import pyqtorch
from pyqtorch import Sequence, Add
from pyqtorch.hamiltonians import Observable

from qadence2_platforms.backends.utils import InputType, Support


def parse_native_observables(observable: list[InputType] | InputType) -> Observable:
    return PyQObservablesParser.build(observable)


class PyQObservablesParser:
    """
    Convert InputType object observables into native PyQTorch object, especially for
    running `expectation` method from PyQTorch interface class. InputType can be
    qadence2-expressions or any other module that implement the same methods.
    """

    @classmethod
    def _add_op(cls, op: InputType) -> Module:
        return Add([cls._get_op(cast(InputType, k)) for k in cast(Iterable, op.args)])

    @classmethod
    def _mul_op(cls, op: InputType) -> Module:
        return Sequence([cls._get_op(cast(InputType, k)) for k in cast(Iterable, op.args)])

    @classmethod
    def _kron_op(cls, op: InputType) -> Module:
        return cls._mul_op(op)

    @classmethod
    def _get_op(cls, op: InputType) -> Primitive | Module:
        """
        Convert an expression into a native PyQTorch object. A simple symbol,
        a quantum operator, and an operation (addition, multiplication or
        kron tensor) are valid objects.

        Args:
            op (InputType): the input expression. Any qadence2-expressions
                expression compatible object or object with same methods.

        Returns:
            A Primitive or torch.nn.Module object.
        """

        if op.is_symbol is True:
            symbol: str = cast(str, op.args[0])
            return getattr(pyqtorch, symbol.upper(), None)

        if op.is_quantum_operator is True:
            primitive: Primitive = cast(Primitive, cls._get_op(cast(InputType, op.args[0])))
            support: Support = cast(Support, op.args[1])

            if support:
                target: int = support.target[0]
                control: list[int] = support.control

                if control:
                    native_op = primitive(target=target, control=control)
                else:
                    native_op = primitive(target=target)

                return native_op

        if op.is_addition is True:
            return cls._add_op(op)

        if op.is_multiplication is True or op.is_kronecker_product is True:
            return cls._mul_op(op)

        raise NotImplementedError(
            f"could not retrieve the expression {op} ({type(op)}) from the observables"
        )

    @classmethod
    def _iterate_over_obs(cls, op: list[InputType] | InputType) -> list[Module]:
        if isinstance(op, Iterable):
            return [cls._get_op(arg) for arg in op]

        args: Iterable = cast(Iterable, op.args)
        return [cls._get_op(arg) for arg in args]

    @classmethod
    def build(cls, observable: list[InputType] | InputType) -> Observable:
        """
        Parses an input expression or list of expressions into a native PyQTorch object.

        Args:
            observable (list[InputType] | InputType): the input expression. Any
                qadence2-expressions expression compatible object or object with same
                methods.

        Returns:
            An PyQTorch Observable object.
        """

        res: list[Module] | Module
        if isinstance(observable, list):
            res = cls._iterate_over_obs(observable)
        else:
            res = [cls._get_op(observable)]
        return Observable(res)
